@@ -32,6 +32,7 @@ For internal use only; no backwards-compatibility guarantees.
 
 import decimal
 import enum
+import functools
 import itertools
 import json
 import logging
@@ -50,7 +51,6 @@ from typing import Set
 from typing import Tuple
 from typing import Type
 
-import dill
 import numpy as np
 from fastavro import parse_schema
 from fastavro import schemaless_reader
@@ -58,6 +58,8 @@ from fastavro import schemaless_writer
 
 from apache_beam.coders import observable
 from apache_beam.coders.avro_record import AvroRecord
+from apache_beam.internal import cloudpickle_pickler
+from apache_beam.internal.cloudpickle import cloudpickle
 from apache_beam.typehints.schemas import named_tuple_from_schema
 from apache_beam.utils import proto_utils
 from apache_beam.utils import windowed_value
@@ -71,8 +73,14 @@ try:
 except ImportError:
   dataclasses = None  # type: ignore
 
+try:
+  import dill
+except ImportError:
+  dill = None
+
 if TYPE_CHECKING:
   import proto
+
   from apache_beam.transforms import userstate
   from apache_beam.transforms.window import IntervalWindow
 
@@ -87,9 +95,9 @@ is_compiled = False
 fits_in_64_bits = lambda x: -(1 << 63) <= x <= (1 << 63) - 1
 
 if TYPE_CHECKING or SLOW_STREAM:
+  from .slow_stream import ByteCountingOutputStream
   from .slow_stream import InputStream as create_InputStream
   from .slow_stream import OutputStream as create_OutputStream
-  from .slow_stream import ByteCountingOutputStream
   from .slow_stream import get_varint_size
 
   try:
@@ -100,10 +108,11 @@ if TYPE_CHECKING or SLOW_STREAM:
 
 else:
   # pylint: disable=wrong-import-order, wrong-import-position, ungrouped-imports
+  from .stream import ByteCountingOutputStream
   from .stream import InputStream as create_InputStream
   from .stream import OutputStream as create_OutputStream
-  from .stream import ByteCountingOutputStream
   from .stream import get_varint_size
+
   # Make it possible to import create_InputStream and other cdef-classes
   # from apache_beam.coders.coder_impl when Cython codepath is used.
   globals()['create_InputStream'] = create_InputStream
@@ -346,6 +355,7 @@ DATACLASS_TYPE = 101
 NAMED_TUPLE_TYPE = 102
 ENUM_TYPE = 103
 NESTED_STATE_TYPE = 104
+DATACLASS_KW_ONLY_TYPE = 105
 
 # Types that can be encoded as iterables, but are not literally
 # lists, etc. due to being lazy.  The actual type is not preserved
@@ -354,14 +364,44 @@ NESTED_STATE_TYPE = 104
 _ITERABLE_LIKE_TYPES = set()  # type: Set[Type]
 
 
+def _verify_dill_compat():
+  base_error = (
+      "This pipeline runs with the pipeline option "
+      "--update_compatibility_version=2.67.0 or earlier. "
+      "When running with this option on SDKs 2.68.0 or "
+      "later, you must ensure dill==0.3.1.1 is installed.")
+  if not dill:
+    raise RuntimeError(base_error + ". Dill is not installed.")
+  if dill.__version__ != "0.3.1.1":
+    raise RuntimeError(base_error + f". Found dill version '{dill.__version__}")
+
+
+dataclass_uses_kw_only: Callable[[Any], bool]
+if dataclasses:
+  # Cache the result to avoid multiple checks for the same dataclass type.
+  @functools.cache
+  def dataclass_uses_kw_only(cls) -> bool:
+    return any(
+        field.init and field.kw_only for field in dataclasses.fields(cls))
+
+else:
+  dataclass_uses_kw_only = lambda cls: False
+
+
 class FastPrimitivesCoderImpl(StreamCoderImpl):
   """For internal use only; no backwards-compatibility guarantees."""
   def __init__(
-      self, fallback_coder_impl, requires_deterministic_step_label=None):
+      self,
+      fallback_coder_impl,
+      requires_deterministic_step_label=None,
+      force_use_dill=False,
+      use_relative_filepaths=True):
     self.fallback_coder_impl = fallback_coder_impl
     self.iterable_coder_impl = IterableCoderImpl(self)
     self.requires_deterministic_step_label = requires_deterministic_step_label
     self.warn_deterministic_fallback = True
+    self.force_use_dill = force_use_dill
+    self.use_relative_filepaths = use_relative_filepaths
 
   @staticmethod
   def register_iterable_like_type(t):
@@ -471,18 +511,25 @@ class FastPrimitivesCoderImpl(StreamCoderImpl):
       self.encode_type(type(value), stream)
       stream.write(value.SerializePartialToString(deterministic=True), True)
     elif dataclasses and dataclasses.is_dataclass(value):
-      stream.write_byte(DATACLASS_TYPE)
       if not type(value).__dataclass_params__.frozen:
         raise TypeError(
             "Unable to deterministically encode non-frozen '%s' of type '%s' "
             "for the input of '%s'" %
             (value, type(value), self.requires_deterministic_step_label))
-      self.encode_type(type(value), stream)
-      values = [
-          getattr(value, field.name) for field in dataclasses.fields(value)
-      ]
+      init_fields = [field for field in dataclasses.fields(value) if field.init]
       try:
-        self.iterable_coder_impl.encode_to_stream(values, stream, True)
+        if dataclass_uses_kw_only(type(value)):
+          stream.write_byte(DATACLASS_KW_ONLY_TYPE)
+          self.encode_type(type(value), stream)
+          stream.write_var_int64(len(init_fields))
+          for field in init_fields:
+            stream.write(field.name.encode("utf-8"), True)
+            self.encode_to_stream(getattr(value, field.name), stream, True)
+        else:  # Not using kw_only, we can pass parameters by position.
+          stream.write_byte(DATACLASS_TYPE)
+          self.encode_type(type(value), stream)
+          values = [getattr(value, field.name) for field in init_fields]
+          self.iterable_coder_impl.encode_to_stream(values, stream, True)
       except Exception as e:
         raise TypeError(self._deterministic_encoding_error_msg(value)) from e
     elif isinstance(value, tuple) and hasattr(type(value), '_fields'):
@@ -525,10 +572,32 @@ class FastPrimitivesCoderImpl(StreamCoderImpl):
         "please provide a type hint for the input of '%s'" %
         (value, type(value), self.requires_deterministic_step_label))
 
+  def encode_type_2_67_0(self, t, stream):
+    """
+    Encode special type with <=2.67.0 compatibility.
+    """
+    if t not in _pickled_types:
+      _verify_dill_compat()
+      _pickled_types[t] = dill.dumps(t)
+    stream.write(_pickled_types[t], True)
+
   def encode_type(self, t, stream):
-    stream.write(dill.dumps(t), True)
+    if self.force_use_dill:
+      return self.encode_type_2_67_0(t, stream)
+
+    if t not in _pickled_types:
+      config = cloudpickle.CloudPickleConfig(
+          id_generator=None,
+          skip_reset_dynamic_type_state=True,
+          filepath_interceptor=cloudpickle.get_relative_path)
+      if not self.use_relative_filepaths:
+        config.filepath_interceptor = None
+      _pickled_types[t] = cloudpickle_pickler.dumps(t, config=config)
+    stream.write(_pickled_types[t], True)
 
   def decode_type(self, stream):
+    if self.force_use_dill:
+      return _unpickle_type_2_67_0(stream.read_all(True))
     return _unpickle_type(stream.read_all(True))
 
   def decode_from_stream(self, stream, nested):
@@ -568,6 +637,14 @@ class FastPrimitivesCoderImpl(StreamCoderImpl):
       msg = cls()
       msg.ParseFromString(stream.read_all(True))
       return msg
+    elif t == DATACLASS_KW_ONLY_TYPE:
+      cls = self.decode_type(stream)
+      vlen = stream.read_var_int64()
+      fields = {}
+      for _ in range(vlen):
+        field_name = stream.read_all(True).decode('utf-8')
+        fields[field_name] = self.decode_from_stream(stream, True)
+      return cls(**fields)
     elif t == DATACLASS_TYPE or t == NAMED_TUPLE_TYPE:
       cls = self.decode_type(stream)
       return cls(*self.iterable_coder_impl.decode_from_stream(stream, True))
@@ -586,22 +663,39 @@ class FastPrimitivesCoderImpl(StreamCoderImpl):
       raise ValueError('Unknown type tag %x' % t)
 
 
+_pickled_types = {}  # type: Dict[type, bytes]
 _unpickled_types = {}  # type: Dict[bytes, type]
 
 
-def _unpickle_type(bs):
+def _unpickle_type_2_67_0(bs):
+  """
+  Decode special type with <=2.67.0 compatibility.
+  """
   t = _unpickled_types.get(bs, None)
   if t is None:
+    _verify_dill_compat()
     t = _unpickled_types[bs] = dill.loads(bs)
     # Fix unpicklable anonymous named tuples for Python 3.6.
     if t.__base__ is tuple and hasattr(t, '_fields'):
       try:
         pickle.loads(pickle.dumps(t))
       except pickle.PicklingError:
-        t.__reduce__ = lambda self: (_unpickle_named_tuple, (bs, tuple(self)))
+        t.__reduce__ = lambda self: (
+            _unpickle_named_tuple_2_67_0, (bs, tuple(self)))
   return t
 
 
+def _unpickle_named_tuple_2_67_0(bs, items):
+  return _unpickle_type_2_67_0(bs)(*items)
+
+
+def _unpickle_type(bs):
+  if not _unpickled_types.get(bs, None):
+    _unpickled_types[bs] = cloudpickle_pickler.loads(bs)
+
+  return _unpickled_types[bs]
+
+
 def _unpickle_named_tuple(bs, items):
   return _unpickle_type(bs)(*items)
 
@@ -837,6 +931,7 @@ class IntervalWindowCoderImpl(StreamCoderImpl):
       if IntervalWindow is None:
         from apache_beam.transforms.window import IntervalWindow
     # instantiating with None is not part of the public interface
+    # pylint: disable=too-many-function-args
     typed_value = IntervalWindow(None, None)  # type: ignore[arg-type]
     typed_value._end_micros = (
         1000 * self._to_normal_time(in_.read_bigendian_uint64()))
@@ -948,7 +1043,14 @@ class VarIntCoderImpl(StreamCoderImpl):
   A coder for int objects."""
   def encode_to_stream(self, value, out, nested):
     # type: (int, create_OutputStream, bool) -> None
-    out.write_var_int64(value)
+    try:
+      out.write_var_int64(value)
+    except OverflowError as e:
+      raise OverflowError(
+          f"Integer value '{value}' is out of the encodable range for "
+          f"VarIntCoder. This coder is limited to values that fit "
+          f"within a 64-bit signed integer (-(2**63) to 2**63 - 1). "
+          f"Original error: {e}") from e
 
   def decode_from_stream(self, in_stream, nested):
     # type: (create_InputStream, bool) -> int
@@ -970,7 +1072,13 @@ class VarIntCoderImpl(StreamCoderImpl):
   def estimate_size(self, value, nested=False):
     # type: (Any, bool) -> int
     # Note that VarInts are encoded the same way regardless of nesting.
-    return get_varint_size(value)
+    try:
+      return get_varint_size(value)
+    except OverflowError as e:
+      raise OverflowError(
+          f"Cannot estimate size for integer value '{value}'. "
+          f"Value is out of the range for VarIntCoder (64-bit signed integer). "
+          f"Original error: {e}") from e
 
 
 class VarInt32CoderImpl(StreamCoderImpl):
@@ -1832,9 +1940,11 @@ class RowCoderImpl(StreamCoderImpl):
       enc_posx = list(
           set(field.encoding_position for field in self.schema.fields))
       if len(enc_posx) != len(self.schema.fields):
+        names_no_pos = ", ".join(
+            [f.name for f in self.schema.fields if f.encoding_position is None])
         raise ValueError(
             f'''Schema with id {schema.id} has encoding_positions_set=True,
-            but not all fields have encoding_position set''')
+            but found fields without encoding_position set: {names_no_pos}''')
       self.encoding_positions = list(
           field.encoding_position for field in self.schema.fields)
     self.encoding_positions_argsort = list(np.argsort(self.encoding_positions))
