@@ -92,12 +92,12 @@ try:
   from bson import json_util
   from bson import objectid
   from bson.objectid import ObjectId
-
   # pymongo also internally depends on bson.
   from pymongo import ASCENDING
   from pymongo import DESCENDING
   from pymongo import MongoClient
   from pymongo import ReplaceOne
+  from pymongo.driver_info import DriverInfo
 except ImportError:
   objectid = None
   json_util = None
@@ -106,6 +106,7 @@ except ImportError:
   DESCENDING = -1
   MongoClient = None
   ReplaceOne = None
+  DriverInfo = None
   _LOGGER.warning("Could not find a compatible bson package.")
 
 __all__ = ["ReadFromMongoDB", "WriteToMongoDB"]
@@ -264,6 +265,12 @@ class _BoundedMongoSource(iobase.BoundedSource):
     self.spec = extra_client_params
     self.bucket_auto = bucket_auto
 
+    if "driver" not in self.spec:
+      self.spec["driver"] = DriverInfo(
+          name="Apache Beam",
+          version=beam.__version__,
+      )
+
   def estimate_size(self):
     with MongoClient(self.uri, **self.spec) as client:
       return client[self.db].command("collstats", self.coll).get("size")
@@ -778,6 +785,12 @@ class _MongoSink:
     self.spec = extra_params
     self.client = None
 
+    if "driver" not in self.spec:
+      self.spec["driver"] = DriverInfo(
+          name="Apache Beam",
+          version=beam.__version__,
+      )
+
   def write(self, documents):
     if self.client is None:
       self.client = MongoClient(host=self.uri, **self.spec)
